/**
     env.c

     shell environment variables handling

     Petr Zika 2002
*/

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "../comdefs.h"
#include "env.h"
#include "../errors/errors.h"
#include "../io/io.h"

/* maximum hash table index */

#define MAXH 1024

/* internal variable structure */
typedef struct variable {
  struct variable *prev;           /* previous */
  struct variable *next;           /* next */
  struct variable *eprev;          /* previous export */
  struct variable *enext;          /* next export */
  char *name;                      /* variable name */
  char *value;                     /* assigned */
  char **export;                   /* exported string && flag in one */
} variable;


LOCAL variable **htable;  /* hash table */
LOCAL variable *exports;  /* exported vars, sharing structure with htable */
LOCAL int export_cnt;     /* size of current environment*/

GLOBAL char **env_list;   /* corresponds to environ */
LOCAL int env_len;        /* env array length !!! with the last NULL entry */


#if 0
/* string duplicate */
LOCAL char *strdup(char *s)
{
   int len = strlen(s) + 1;
	char *t = (char *)malloc(sizeof(char)*len);

	if (!t) return NULL;

	return strcpy(t,s);
}
#endif

/** initialize ds for variables */
GLOBAL int init_vars(char **env)
{
   int i, n;
   char **e;
	char *s, *t;

   if (!(htable = (variable **)malloc(sizeof(variable *)*MAXH)))
      return 1;

   for (i = 0; i < MAXH; i++) htable[i] = NULL;
   
   /* no exported variables so far */
   exports = NULL;
	export_cnt = 0;
   
   /* count own environment variables, add one for NULL */
   n = 1;
   for (e = env; *e; e++) n++;
   
   env_len = (n < 64) ? 64 : n;  /* estimate on exported variable count */

   if (!(env_list = (char **)malloc(sizeof(char *)*env_len))) {
      free(htable);
      error("env: malloc error");
      return 1;
   }
 
   for (e = env; *e; e++) {
		 if ((s = strdup(*e)) == NULL) {
          /* clean up */
			 done_vars();
          error("env: malloc error");
			 return 1;
		 }
		
		 /* split varname and value */ 
		 if ((t = strchr(s,'=')) != NULL) {
			 *t = '\0';
			 t++;
			 if (*t == '\0') t = NULL;  /* empty value */
		 }

		 /* internally add variable to all structures */
		 if (set_variable(s,t) || export_variable(s)) {
			 done_vars();
			 free(s);
          error("env: init error");
			 return 1;
		 }
		 free(s);
   }

   return 0;
}

/** cleanup for variables */
GLOBAL int done_vars(void)
{
	int i;
	variable *u, *v;
	char **e;

   for (i = 0; i < MAXH; i++) {
      if (htable[i]) {
			/* free linklist */
         for (u = htable[i], v = u->next; v; u = v, v = v->next) {
             /* if entry == NULL, free does nothing */
 				 free(u->name);
				 free(u->value);
				 free(u);
			}
	   }
   }
	
	/* the table itself was malloced as well */
	free(htable);

	/* release environment strings */
   for (e = env_list; *e; e++)
       free(*e);

   /* free environment */
   free(env_list);
   
	return 0;
}

/** compute hash function for non NULL nam */
LOCAL int hash_fun(char *nam)
{
   int h = 0;

	for (; *nam; nam++) h = ((h << 5) + (h << 2) + *nam) % MAXH;
	return h % MAXH;
}

/** find certain variable */
LOCAL variable *find(char *nam)
{
   int h;
	variable *v;
  
	h = hash_fun(nam);
	v = htable[h];

	if (!v) return NULL;

	for (; v; v = v->next)
       if (!strcmp(v->name,nam)) return v;

   return NULL;
}

LOCAL int check_var_name(char *nam)
{
	char c;

	if (!nam) return 1;

	while ((c = *nam) != '\0') {
		if (!isalpha(c) && !isdigit(c) && c != '_') return 1; 
      nam++;
	}

   return 0;
}

/** set variable to value */
GLOBAL int set_variable(char *nam, char *val)
{
   variable *v = find(nam);
	int len, i, h;
	char *s, *t;

	if (val && *val == '\0') val = NULL;
   if (check_var_name(nam)) return 1;

   /* create nonexistent */
	if (!v) {
      if (!(v = (variable *)malloc(sizeof(variable))))
			return 1;
		if ((v->name = strdup(nam)) == NULL) {
			free(v);
			return 1;
		}

		if (val) {
		   if ((s = strdup(val)) == NULL) {
			   free(v);
				return 1;
			}
		} else s = NULL;
		
		v->value = s;
		v->prev = NULL;
      v->export = NULL;

		/* insert new record to front*/
		h = hash_fun(nam);
		v->next = htable[h];
		if (htable[h]) htable[h]->prev = v;
		htable[h] = v;

	} else {

		/* no change in value */
		if ((!val && !v->value) || (val && v->value && !strcmp(val,v->value)))
			return 0;

		/* set new value */
		if (val) {
			if ((s = strdup(val)) == NULL)
				return 1;
		} else s = NULL;
			
		/* alter export string */
		if (v->export) {
			
			/* create new environment entry */
			i = len = strlen(v->name);
			if (s) len += strlen(s);
			len++;
			
			if ((t = (char *)malloc(len + 1)) == NULL) {
				free(s);
				return 1;
			}

			strcpy(t,v->name);
			t[i++] = '=';
			if (s) 
				strcpy(&t[i],s);
			else t[i] = '\0';

			free(*(v->export));
			*(v->export) = t;
		}
		
		free(v->value);
		v->value = s;
	}
   
		
	return 0;
}

/** get value of variable */
GLOBAL char *get_variable(char *nam)
{
   variable *v = find(nam);
	
	if (!v) return NULL;
	return v->value;
}

/** prepare variable for export */
GLOBAL int export_variable(char *nam)
{
   variable *v = find(nam);
	int i, len;
	char **e;
	char *s;
  
   if (!v) return 1;
   if (!v->export) {
		
		export_cnt++;

		/* enlarge base array */
      if (export_cnt + 1 > env_len) {
			len = 2*(env_len - 1) + 1;  /* double storage, add NULL */
			e = (char **)realloc(env_list,sizeof(char *)*len);
			if (!e) {
				len = env_len + 1;       /* increase storage */
			   e = (char **)realloc(env_list,sizeof(char *)*len);
				if (!e) {
					v->export = NULL;
					export_cnt--;
					return 1;
				}
			}
         env_list = e;
			env_len = len;
		}
		
		/* create new environment entry */
		i = len = strlen(v->name);
		if (v->value) len += strlen(v->value);
      len++;
		
		if ((s = (char *)malloc(len + 1)) == NULL) {
			v->export = NULL;
			export_cnt--;
			return 1;
		}

		strcpy(s,v->name);
		s[i++] = '=';
		if (v->value) 
			strcpy(&s[i],v->value);
		else s[i] = '\0';
		
		if (exports) {
			/* insert into front */
			exports->eprev = v;
			v->eprev = NULL;
			v->enext = exports;
			exports = v;
			
			/* insert into back of array */
			env_list[export_cnt - 1] = s;
			env_list[export_cnt] = NULL;
         v->export = env_list + export_cnt - 1;
			
		} else {
			v->eprev = v->enext = NULL;
			exports = v;
			env_list[0] = s;
			env_list[1] = NULL;
         v->export = env_list;
		} 
	}
   return 0;	
}

/** declare variable local, delete export flag */
GLOBAL int localize_variable(char *nam)
{
	variable *u;
   variable *v = find(nam);
	char **e;
  
   if (!v) return 1;
   if (v->export) {
		export_cnt--;

		if (v == exports) {
			/* no predecessor, discard head */
			exports = exports->enext;
			exports->eprev = NULL;

			/* free the last env_list entry */
			free(env_list[export_cnt]);
			env_list[export_cnt] = NULL;
			
		} else {

			/* move array to front, including NULL */
         e = v->export;
			free(*e);
			
			u = v;
			while (u) {
				u = u->eprev;
				*e = *(e + 1);
				e++;
			}
		
			v->eprev->enext = v->enext;
		   if (v->enext) v->enext->eprev = v->eprev;
		}
				
      /* do not set v->enext,eprev, not used until v->export */
		v->export = NULL;


      /* TODO reallocate exports! */
	}
   return 0;	
}

/** delete variable from structures */
GLOBAL int unset_variable(char *nam)
{
   variable *u;
   variable *v = find(nam);
	int h;
  
   if (!v) return 1;

	localize_variable(nam); /* unlink from env_list structures */

	u = v->prev;

	if (u) {
		/* non NULL predessor, links easily */
      u->next = v->next;
	} else {
      /* u is the first in the list */
		h = hash_fun(nam);
		htable[h] = u;
	}
	
	if (v->next) v->next->prev = u;

	free(v);

	return 0;
}

/** get environment */
GLOBAL char **get_env(void)
{
	return env_list;
}

/** print environment */
GLOBAL void  print_env(void)
{
	char **e = env_list;

   while (*e) {
     printf("%s\n",*e);
	  e++;
	}
}

GLOBAL void print_local(void)
{
	variable *v;
	int i;

	for (i = 0; i < MAXH; i++) {
		v = htable[i];

		while (v) {
         if (!v->export)
            printf("%s=%s\n",v->name,(v->value == NULL)?"":v->value);
			v = v->next;
		}
	}
}

#ifdef TEST_ENV

int main(void)
{
	char *u;
	char **e;
	
	init_vars();

   set_variable("abc","def");
   export_variable("abc");
   set_variable("ghi","jkl");
   export_variable("ghi");
	
   u = get_variable("ghi");
	printf("%s=%s\n","ghi",u);
	u = get_variable("abc");
	printf("%s=%s\n","abc",u);
  	 
   set_variable("abc",NULL);
	u = get_variable("abc");
	printf("%s=%s\n","abc",u);
    	
   set_variable("mno","pqr");
   export_variable("mno");
	u = get_variable("mno");
	printf("%s=%s\n","mno",u);

   localize_variable("ghi");
	
	if (!(e = get_env())) {
		printf("error getting env\n");
	} else {
     printf("===============================================\n");
     print_env(e);
	}

	done_vars();
	
   return 0;
}

#endif

