/**
     grbuild.y
		  
     build and process input grammar
				 
	  Petr Zika 2002
*/

%{
#include <string.h>
#include <stdlib.h>
#include <setjmp.h>
#include "../comdefs.h"
#include "grbuild.h"
#include "../tree/tree.h"
#include "../lexer/lexer.h"
#include "../errors/errors.h"
#include "../utils/utils.h"
#include "../env/env.h"
#include "../run/run.h"
#include "../io/io.h"
#include "../main/main.h"

/* pointers to structures for error recovery deallocation */
LOCAL cmd_list *save_cmds;
LOCAL arg_list *save_args;
LOCAL redir_list *save_redirs;

LOCAL int should_run;  /* flag for next command */
LOCAL int last_return; /* last cmd_list return value */
LOCAL int leaving;     /* exit called */

%}

%union {
   char *text;                 /* any string */
   char type;                  /* redirection and list separator type */ 
	cmd_list *cmds;             /* commands structure */
	arg_list *args;             /* arguments structure */
	redir_list *redirs;         /* redirections structure */
}

%type <cmds> command
%type <cmds> cmd_list
%type <args> arg
%type <args> arg_list
%type <redirs> redir
%type <redirs> redir_list
%type <type> redir_type
%type <text> name
%type <type> list_separator

%token <type> SYN_IN    300
%token <type> SYN_OUT   301
%token <type> SYN_2OUT  302
%token <type> SYN_APP   303
%token <type> SYN_2APP  304

%token SYN_PIPE         305

%token <type> SYN_OR    307
%token <type> SYN_AND   308

%token SYN_SEMICOLON    306
%token SYN_EOLN         309

%token SYN_ERROR        310

%token <text> SYN_WORD  311
%token SYN_EQ           312

%%

/* whole input are virtual "command lines" */
input: command_line {} ;
input: input command_line {} ;
   			

/* error recovery at the end of command line */
command_line: error line_sep 
{
#ifdef TEST_GRBUILD
   error("recovering from error 2");
#endif

   should_run = 1;
   last_return = 0;
	
	free_cmd_list(save_cmds);
	free_arg_list(save_args);
	free_redir_list(save_redirs);
	save_cmds = NULL;
	save_redirs = NULL;
	save_args = NULL;
	yyerrok;
} ;

/* command line is formed by a list separated by line separator */
command_line: line_sep {} ;
command_line: list line_sep 
{
   /* independent list of commands */
   should_run = 1;
	last_return = 0;
} ;

/* lines are separated by semicolons, end of line terminates the last line */
line_sep: SYN_EOLN {} ;
line_sep: SYN_SEMICOLON {} ;

/* put the error through */
list: error
{
#ifdef TEST_GRBUILD
   error("recovering from error 3");
#endif
} ;

/* recovering from an error inside a list */
list: error list_separator
{
#ifdef TEST_GRBUILD
   error("recovering from error 4");
#endif

   last_return = 1; /* force error state */
	
   if ($2 == SEP_AND)
	    should_run = 0;
	else
	    should_run = 1;

	free_cmd_list(save_cmds);
	free_arg_list(save_args);
	free_redir_list(save_redirs);
	save_cmds = NULL;
	save_redirs = NULL;
	save_args = NULL;

	yyerrok;
}     
/* continue with the next list */
      list {} ;

/* list are list items and list separators in a sequence 
   rule is intentionally right recursive, due to partial evaluation */
list: list_item ;
list: list_item list_separator
{
	/* && forces processing the next list_item if previous was OK */
   if ($2 == SEP_AND)
	   should_run = (last_return == 0);
	else
	   should_run = (last_return != 0);
}
/* continue with the next list */
    list {} ;


/* list item is a variable assignment or a command list */
list_item: assignment ;
list_item: cmd_list
{
   int status;
	cmd_list *c = $1;

   prune_cmd_list(c);   /* clear ambiguous redirects */

#ifdef TEST_GRBUILD
   print_cmd_list(c);
   putchar('\n');
#endif

	/* flag set by conditional separators */
   if (should_run) {
      is_running = 1;
		status = run_cmd_list(c);
      is_running = 0;

      if (status == -1)
			last_return = 1;
		else if (status == -2)
         leaving = 1;
      else 
		   last_return = status;
	} else should_run = 1;

	free_cmd_list(c);
	save_cmds = NULL;
   if (leaving) zzerror("");
} ;

/* assignment may be ill formed, with leading word */
assignment: SYN_WORD SYN_EQ SYN_WORD SYN_WORD
{
	error("ill formed assignment `%s=%s %s...'",$1,$3,$4);
   YYERROR;  /* force error to eat the rest of the arguments */
} ;

/* assignment with no value  */
assignment: SYN_WORD SYN_EQ
{
   if (set_variable($1,NULL)) {
      error("could not set variable `%s'",$1);
      last_return = 1;
   }
} ;

/* ordinary assignment with both sides */
assignment: SYN_WORD SYN_EQ SYN_WORD
{
   if (set_variable($1,$3)) {
      error("could not set variable `%s'",$1);
      last_return = 1;
   }
} ;

/* ill formed command list with no command before pipe */
cmd_list: SYN_PIPE
{
	error("no command before pipe");
   YYERROR;
} ;

/* ill formed command list with no command after pipe */
cmd_list: cmd_list SYN_PIPE
{
	error("no command after pipe near `%s'",$1->last->name);
   YYERROR;
} ;

/* command list may be either single command or pipeline */
cmd_list: command ;
cmd_list: cmd_list SYN_PIPE command
{
   cmd_list *c, *d;

	c = $1;
	d = $3;
	c->last->next = d;
	c->last = d;
	$$ = c;
} ;

/* bare command */
command: name
{
   cmd_list *c;

	c = NEW(cmd_list);
	c->name = $1;
	c->args = NULL;
	c->redirs = NULL;
	c->next = NULL;
	c->last = c;
   save_cmds = c;
	$$ = c;
} ;

/* command with arguments */
command: name arg_list
{
   cmd_list *c;

	c = NEW(cmd_list);
	c->name = $1;
	c->args = $2;
	c->redirs = NULL;
	c->next = NULL;
	c->last = c;
	save_args = NULL;
   save_cmds = c;
	$$ = c;
} ;

/* command with redirections */
command: name redir_list
{
   cmd_list *c;

	c = NEW(cmd_list);
	c->name = $1;
	c->args = NULL;
	c->redirs = $2; 
	c->next = NULL;
	c->last = c;
	save_redirs = NULL;
   save_cmds = c;
	$$ = c;
} ;

/* command with arguments and redirections */
command: name arg_list redir_list
{
   cmd_list *c;

	c = NEW(cmd_list);
	c->name = $1;
	c->args = $2;
	c->redirs = $3;
	c->next = NULL;
	c->last = c;
	save_redirs = NULL;
	save_args = NULL;
   save_cmds = c;
	$$ = c;
} ;

/* argument list is a list of one or more arguments */
arg_list: arg ;
arg_list: arg_list arg
{
   arg_list *a, *b;

	a = $1;
	b = $2;
	a->last->next = b;
	a->last = b;
	$$ = a;
} ;

/* argument is a single word */
arg: SYN_WORD
{
   arg_list *a;

	a = NEW(arg_list);
	a->name = $1;
	a->next = NULL;
	a->last = a;
   save_args = a;
	$$ = a;
} ;

/* name is a word as well */
name: SYN_WORD ;

/* redirection list is a list of redirections */
redir_list: redir ;
redir_list: redir_list redir
{
   redir_list *r = $1;
	redir_list *s = $2;

	if (r != NULL) {
	   if (s != NULL) {
		   r->last->next = s;
	      r->last = s;
		} /* else nothing to add */
	} else if (s != NULL) {
	  /* list is empty so far, make s the first entry */
     r = s;
	  save_redirs = r;
	} /* else both NULL, do nothing */
	
   $$ = r;
} ;

/* illformed redirection without the file part */
redir: redir_type
{
	error("indefinite redirection");
	YYERROR; 
} ;

/* redirection consists of redirection type token and a file name */
redir: redir_type SYN_WORD
{
   redir_list *r;
   char *c = $2;

   if (c != NULL) {
		r = NEW(redir_list);
		r->type = $1;
		r->name = c;
		r->next = NULL;
		r->last = r;
		save_redirs = r;
	} else r = NULL;
	
	$$ = r;
} ;

/* possible redirection types */
redir_type: SYN_IN ;
redir_type: SYN_OUT ;
redir_type: SYN_2OUT ;
redir_type: SYN_APP ;
redir_type: SYN_2APP ;

/* list separators are conditional */
list_separator: SYN_OR ;
list_separator: SYN_AND ;

%%

/** bison token parser */
GLOBAL int zzlex(void)
{
   int i = read_lex();

	if (i == SYN_ERROR)
      zzerror("lexer: token error");
   else if (i == SYN_WORD)
		zzlval.text = STRDUP(lex_word);
	else if (SYN_IN <= i && i <= SYN_2APP)
      /* this IS dependent on the constants! */
		zzlval.type = i - SYN_IN + RED_IN;
	else if (i == SYN_AND || i == SYN_OR)
		zzlval.type = i - SYN_AND + SEP_AND;

   return i;
}

/** bison error recovery routine */
GLOBAL void zzerror(char *s)
{
   if (!leaving) error("parser: %s\n",s);

	free_cmd_list(save_cmds);
	free_arg_list(save_args);
	free_redir_list(save_redirs);
   
   return_handler(1);
}

/** build and process grammar
    only catches error which could not be handled statically by grcheck
*/
GLOBAL int grammar_build(char *line)
{
   int err = 0;

   lex_state = LEX_BUILD;  /* building, so actually expand everything */

	save_cmds = NULL;
	save_args = NULL;
	save_redirs = NULL;

	should_run = 1;
	last_return = 0;
	leaving = 0;

   /* set new return handler address, for MALLOC etc. */
   if ((err = create_handler()) == 0) {
      if (init_lex_parser(line)) {
         destroy_handler(); 
         return -1;
      }
      if (zzparse()) err = 1;
   } else {
     /* distinguish handler error, exit call and error return */
     if (err != -1) err = 1;
     if (leaving) err = -2;
   } 

	/* handler is local to this function */
   if (err != -1) {
      destroy_handler();
      if (done_lex_parser()) return -1;
	} else
	   /* unable to create handler is error */	
		err = 1;

   return err;
}

#ifdef TEST_GRBUILD

char *read_line(char *s, int max)
{
	int l;
	if (!fgets(s,max,stdin)) return NULL;

	l = strlen(s);
	if (l && s[l - 1] == '\n') s[l - 1] = '\0';

   return s;
}

extern char **environ;

int main(void)
{
   char line[MAX_LINE + 1];
   int y;

   if (init_vars(environ)) exit(1);

   do {
      printf("grbuild:\n");

      if (!read_line(line,MAX_LINE)) return 0;
      printf("line just read: \"%s\"\n",line);
    
      y = grammar_build(line);

      switch (y) {
			case -1:
				printf("internal error!\n");
				break;
			case 0:
				printf("line OK\n");
				break;
			default:
				printf("grammar error!\n");
				break;
      }

   } while (1);

   if (done_vars()) exit(1);

   return 0;
}

#endif

