/**
     grcheck.y
		  
     check consistency of grammar
				 
	  Petr Zika 2002
*/
      
%{
#include <string.h>
#include <stdlib.h>
#include <setjmp.h>
#include "../comdefs.h"
#include "grcheck.h"
#include "../lexer/lexer.h"
#include "../errors/errors.h"
#include "../io/io.h"
%}

/* WARNING: this file has to be manually updated to match 
   the grammar present in file grbuild.y 
   the constants depend on the LEX_ constants in ../lexer/lexer.c
*/

%token SYN_IN    300
%token SYN_OUT   301
%token SYN_2OUT  302
%token SYN_APP   303
%token SYN_2APP  304

%token SYN_PIPE  305

%token SYN_OR    307
%token SYN_AND   308

%token SYN_SEMICOLON    306
%token SYN_EOLN  309

%token SYN_ERROR 310

%token SYN_WORD  311
%token SYN_EQ    312

%%

/* whole input are virtual "command lines" */
input: command_line {} ;
input: input command_line {} ;

/* command line is formed by a list separated by line separator */
command_line: line_sep {} ;
command_line: list line_sep {} ;

/* lines are separated by semicolons, end of line terminates the last line */
line_sep: SYN_EOLN {} ;
line_sep: SYN_SEMICOLON {} ;

/* list are list items and list separators in a sequence 
   rule is intentionally right recursive, due to partial evaluation */
list: list_item {} ;
list: list_item list_separator list {} ;


/* list item is a variable assignment or a command list */
list_item: assignment {} ;
list_item: cmd_list {} ;

/* assignment with no value  */
assignment: SYN_WORD SYN_EQ {} ;

/* ordinary assignment with both sides */
assignment: SYN_WORD SYN_EQ SYN_WORD {} ;


/* command list may be either single command or pipeline */
cmd_list: command {} ;
cmd_list: cmd_list SYN_PIPE command {} ;

/* bare command */
command: name {} ;

/* command with arguments */
command: name arg_list {} ;

/* command with redirections */
command: name redir_list {} ;

/* command with arguments and redirections */
command: name arg_list redir_list {} ;

/* argument list is a list of one or more arguments */
arg_list: arg {} ;
arg_list: arg_list arg {} ;

/* argument is a single word */
arg: SYN_WORD {} ;

/* name is a word as well */
name: SYN_WORD {} ;

/* redirection list is a list of redirections */
redir_list: redir {} ;
redir_list: redir_list redir {} ;

/* redirection consists of redirection type token and a file name */
redir: redir_type SYN_WORD {} ;

/* possible redirection types */
redir_type: SYN_IN {} ;
redir_type: SYN_OUT {} ;
redir_type: SYN_2OUT {} ;
redir_type: SYN_APP {} ;
redir_type: SYN_2APP {} ;

/* list separators are conditional */
list_separator: SYN_OR {} ;
list_separator: SYN_AND {} ;

%%

/* last word for error reporting */
LOCAL char last_word[MAX_LINE + 1];

/* bison token parser */
GLOBAL int yylex(void)
{
   int i = read_lex();

	if (i == SYN_ERROR)
		yyerror("lexer: token error");
   else if (i == SYN_WORD) 
      strcpy(last_word,lex_word);

   return i;
}

/* bison error recovery routine */
GLOBAL void yyerror(char *s)
{
   if (*last_word) 
      error("parser: %s near `%s'\n",s,last_word);
   else 
      error("parser: %s\n",s);

   return_handler(1);
}

/** check grammar corectness, return 1 for incorrect */
GLOBAL int grammar_check(char *line)
{
   int err = 0;

   lex_state = LEX_CHECK;
   *last_word = '\0';

   if ((err = create_handler()) == 0) {
      if (init_lex_parser(line)) {
         destroy_handler();
         return -1;
      }
      if (yyparse()) err = 1;
   } else {
	   /* distinguish between handler and return error */	
      if (err != -1) err = 1;
	}
		
	if (err != -1) {
		destroy_handler();
      if (done_lex_parser()) return -1;
	} else
	   /* unable to create handler */	
		err = 1;

   return err;
}

#ifdef TEST_GRCHECK

char *read_line(char *s, int max)
{
	int l;
	if (!fgets(s,max,stdin)) return NULL;

	l = strlen(s);
	if (l && s[l - 1] == '\n') s[l - 1] = '\0';

   return s;
}

int main(void)
{
   char line[MAX_LINE + 1];
   int y;
  
   do {
      printf("grcheck:\n"); 

      if (!read_line(line,MAX_LINE)) return 0;
      printf("line just read: \"%s\"\n",line);
      
      
      y = grammar_check(line);

      switch (y) {
			case -1:
				printf("internal error!\n");
				break;
			case 0:
				printf("line OK\n");
				break;
			default:
				printf("grammar error!\n");
				break;
      }

   } while (1);

   return 0;
}

#endif

