/**
     io.c

     input/output primitives

     Petr Zika 2002
*/

/*
#define _ISOC99_SOURCE
*/

#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include "io.h"
#include "../comdefs.h"
#include "../errors/errors.h"

#define MAX_BUF 1024

extern int vsnprintf(char *str, size_t size, const char *format, va_list ap);

LOCAL char buf[MAX_BUF];

/** print formatted string to stdin */
GLOBAL int printf(char *str, ...)
{
	va_list v;
	int len = MAX_BUF;
	int n;
	char *s = buf;
	char *t = NULL;

	va_start(v,str);
	n = vsnprintf(buf,MAX_BUF,str,v);
	va_end(v);
	
	while (1) {
		if (n > -1 && n < len)
			break;
		
		free(t);  /* first time free(NULL) */

		if (n > -1) 
			len = n + 1;
		else 
			len *= 2;
		
		if ((t = (char *)malloc(sizeof(char)*len)) == NULL)
				fatal("printf: not enough memory");

		s = t;
		
		va_start(v,str);
		n = vsnprintf(buf,len,str,v);
		va_end(v);
   }

   if (write(stdout,s,strlen(s)) == -1)
      fatal("fprintf: i/o error"); 

   free(t);	  /* could be NULL as well */
	
	return n - 1;
}

/** print formatted string to file */
GLOBAL int fprintf(int fd, char *str, ...)
{
	va_list v;
	int len = MAX_BUF;
	int n;
	char *s = buf;
	char *t = NULL;

	va_start(v,str);
	n = vsnprintf(buf,MAX_BUF,str,v);
	va_end(v);
	
	while (1) {
		if (n > -1 && n < len)
			break;
		
		free(t);  /* first time free(NULL) */

		if (n > -1) 
			len = n + 1;
		else 
			len *= 2;
		
		if ((t = (char *)malloc(sizeof(char)*len)) == NULL)
         /* self recursion here not possible, printing in local buf */ 
			fatal("fprintf: not enough memory");

		s = t;
		
		va_start(v,str);
		n = vsnprintf(buf,len,str,v);
		va_end(v);
   }

   if (write(fd,s,strlen(s)) == -1) {
      if (fd != stderr) fatal("fprintf: i/o error"); /* to avoid recursion */
	   die(); /* not being able to write to stderr is serious */
   }

   free(t);	
	
	return 0;
}

LOCAL char char_buf; 

/** get single char */
GLOBAL int fgetc(int fd)
{
	int res;
   if ((res = read(fd,&char_buf,1)) == -1)
      return EOF;

	if (!res) return EOF;

   return char_buf;	
}

/** put single char */
GLOBAL int fputc(int c, int fd)
{
   int res;

	char_buf = (unsigned char)c;
	if ((res = write(fd,&char_buf,1)) == -1)
		fatal("fputc: i/o error");

   return (int)char_buf;	
}

/** get whole string */
GLOBAL char *fgets(char *s, size_t size, int fd)
{
   int i;
	int c;
	char *t = s;

	i = 1;
	while (i < size && (c = fgetc(fd)) != EOF) {
		*s++ = (char)c;
		if (c == '\n') break;
		i++;
	}
   *s = '\0';

   if (c == EOF) {
      if (errno || t == s) return NULL;
   }
   return t;
}

/** put whole string */
GLOBAL int fputs(const char *s, int fd)
{
	if (write(fd,s,strlen(s)) == -1)
		fatal("fputs: i/o error");

	return 1;
}

#ifdef TEST_IO

int main(void)
{
    char str[20];
	 
	 printf("test %d %c %s\n",42,'x',"printf");
	 fprintf(stdout,"test stdout %d %c %s\n",42,'x',"fprintf");
	 fprintf(stderr,"test stderr %d %c %s\n",42,'x',"fprintf");

	 putchar('a');
	 putchar('\n');

	 fputs("Enter string:",stdout);

	 fgets(str,20,stdin);
	 printf("You entered: %s\n",str);

    return 0;
}

#endif

#undef MAX_BUF

