/**
       lexer.c
 
       lexical parser
 
       Petr Zika 2002
*/
                 
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <ctype.h>
#include "../comdefs.h"
#include "lexer.h"
#include "../expand/expand.h"
#include "../env/env.h"

/* state of lexer, checking or building 
   this generally means, whether to expand wildcards and variables or not 
   in static checking state within grcheck, we suppose every expansion/variable
   to be "nice" in the way that it does not create mess in the input grammar
   hopefully grbuild catches the rest of the cases not covered by static check
*/

GLOBAL int lex_state;  

/* in here messing up different things
   token constants are both for low level signalling
   and high level word flags */

typedef enum { 
  TOK_CHAR,       /* any ordinary char */
  TOK_DOLAR,      /* $ */
  TOK_DOLAR2,     /* $ inside dquote */
  TOK_LT,         /* < */
  TOK_LT2,        /* < inside dquote */ 
  TOK_GT,         /* > */
  TOK_GT2,        /* > inside dquote */
  TOK_SEMICOLON,  /* ; */
  TOK_OR,         /* | */
  TOK_ET,         /* & */
  TOK_QMARK,      /* ? */
  TOK_STAR,       /* * */
  TOK_EQ,         /* = */
  TOK_BLANK,      /* space, tab */
  TOK_EOLN,       /* end of line (end of input) */
  TOK_ERROR,      /* error flag */
  TOK_EMPTY,      /* "" or '' as empty word */
  TOK_NONE,       /* for dquote end signalling */
  TOK_TWO,        /* single character '2' */
  TOK_UNDEF,      /* word parser flag */
  TOK_WORD,       /* word parser flag */
  TOK_NOVAR       /* variable expanded to empty string */
} token;

/* word structure 
   if tok == TOK_WORD, text is pointing into buffer */
typedef struct {
   char *text;
	token tok;
} word;


/* bit manipulation macros */
#define TOK_CHAR_FLAG (1 << CHAR_BIT)
#define MAKE_TOK_CHAR(c) (c | TOK_CHAR_FLAG)
#define IS_TOK_CHAR(t) (t & TOK_CHAR_FLAG)
#define GET_TOK_CHAR(t) (t & ~TOK_CHAR_FLAG)

LOCAL char next_char;       /* next character in buffer */
LOCAL char *curr_char;      /* pointer into buffer */
LOCAL int in_quote;         /* flag for quoted strings */
LOCAL int in_dquote;         /* flag for quoted strings */
LOCAL int empty_quote;      /* flag for empty quotes */

/** initialize variables used for token parsing */
LOCAL int init_token_parser(char *s)
{
   /* no buffer passed */
   if (!s) return -1;

   curr_char = s;
   next_char = *curr_char;
   in_quote = 0;
   return 0;
}

/** finish token parsing */
LOCAL int done_token_parser(void) 
{
   return 0;
}

/** skip next character in input */
LOCAL void skip_char(void)
{
    if (!next_char) return;

    curr_char++;
    next_char = *curr_char;
}

/** read next character in input */
LOCAL int read_char(void)
{
    char tmp;
    /* stop at the end of the string */
    if (!next_char) return 0;

    tmp = next_char;
    /* advance pointer */
    curr_char++;
    next_char = *curr_char;
    
    return tmp;
}

/** read one token and return its value */
LOCAL token read_token(void)
{
   token t;
	char c;

read_token_start:

   if (in_quote) {
     /* in quoted string */
     if (next_char == '\'') {
        /* leaving quote */
        skip_char();
        in_quote = 0;
		  /* '' is explicit empty string */
		  if (empty_quote) t = TOK_EMPTY;
        /* if no output generated, get some */
        else goto read_token_start;

     } else if (!next_char)
       /* unfinished quote */
       t = TOK_ERROR;
     else 
       /* create ordinary character */
       t = MAKE_TOK_CHAR(read_char());
	  empty_quote = 0;
     
   } else if (in_dquote) {
     /* almost nothing allowed in doublequote, except $ < > \.
        all other features treated as normal char */
		switch (next_char) {
			case '\0':
				t = TOK_ERROR;
				break;
			case '\"':
				/* exit dquoted string */
				skip_char();
				in_dquote = 0;
				if (empty_quote) 
					t = TOK_EMPTY;
				else 
					/* break word at the end of "..." */
					t = TOK_NONE;
				break;
			case '$':
				skip_char();
				t = TOK_DOLAR2;
				break;
			/* lt/gt allowed only as $<name> */
			case '<':
				skip_char();
				t = TOK_LT2;
				break;
			case '>':
				skip_char();
				t = TOK_GT2;
				break;
			case '\\':
				/* escape character reached */
				skip_char();
				if (!(c = read_char())) 
					/* splitting lines is not allowed */
					t = TOK_ERROR; 
				else 
					/* return just any character as ordinary */
					t = MAKE_TOK_CHAR(c);
				break;
			default:
				t = MAKE_TOK_CHAR(read_char());
				break;
			
		}
		/* already something in quote */
      empty_quote = 0;
           
   } else { 
      /* out of (d)quoted string */
		switch (next_char) {
		  case '#':
           /* skip comment */
			  while (next_char) skip_char();
			  t = TOK_EOLN;
			  break;
        case '\0':
           t = TOK_EOLN;
           break;
		  case ' ':
		  case '\t':
           /* assume noone wants to distinguish tab from space */
			  skip_char();
			  t = TOK_BLANK;
			  break;
		  case '\'':
           /* entering quote - see first if branch */
			  skip_char();
			  in_quote = 1;
			  empty_quote = 1;
           goto read_token_start;
		  case '\"':
           /* entering dquote - see second if branch */
           skip_char();
		     in_dquote = 1;
			  empty_quote = 1; /* same flag as for quote */
           goto read_token_start;
			  break;
        case '\\':
           /* escape character reached */
			  skip_char();
			  if (!(c = read_char())) 
				  /* splitting lines is not allowed */
 				  t = TOK_ERROR; 
			  else 
				  /* return just any character as ordinary */
				  t = MAKE_TOK_CHAR(c);
           break;
		  /* other special chars are all the same */
		  case '$':
			  skip_char();
			  t = TOK_DOLAR;
			  break;
		  case '<':
			  skip_char();
			  t = TOK_LT;
			  break;
		  case '>':
			  skip_char();
			  t = TOK_GT;
			  break;
		  case ';':
			  skip_char();
			  t = TOK_SEMICOLON;
			  break;
		  case '|':
			  skip_char();
			  t = TOK_OR;
			  break;
		  case '&':
			  skip_char();
			  t = TOK_ET;
			  break;
		  case '?':
			  skip_char();
			  t = TOK_QMARK;
			  break;
 		  case '*':
			  skip_char();
			  t = TOK_STAR;
			  break;
 		  case '=':
			  skip_char();
			  t = TOK_EQ;
			  break;
        default:
			  /* ordinary character */
			  t = MAKE_TOK_CHAR(read_char());
           break;
      }
   }
   return t;
}

LOCAL int is_var;                         /* variable flag */
LOCAL int was_lt;                         /* < flag */
LOCAL char word_buffer[2*(MAX_LINE + 1)]; /* buffer for joined chars */
LOCAL char *word_buf_ptr;                 /* current position in buffer */
LOCAL token old_token;                    /* token before current */
LOCAL int was_word;

/** initialize word parser */
LOCAL int init_word_parser(void)
{
	word_buf_ptr = word_buffer;
	is_var = 0;
	was_lt = 0;
   old_token = TOK_UNDEF;
	was_word = 0;
	
   return 0;
}

/** finish word parser */
LOCAL int done_word_parser(void)
{
   return 0;
}

/** read one word from input, return in structure */
LOCAL word read_word(void)
{
	word wrd;
	token tok;
	char *text_start;
   char c;

read_word_start:

	wrd.text = NULL;
	text_start = word_buf_ptr;

   /* perhaps some preread token from last time */
	if (old_token == TOK_UNDEF)
		tok = read_token();
	else { 
		tok = old_token;
      /* yet no preread token */
	   old_token = TOK_UNDEF;
	}

	if (is_var) {
		if (IS_TOK_CHAR(tok)) {
		   c = GET_TOK_CHAR(tok);
			/* check variable name starts with proper character */
			if (!isalpha(c) && !isdigit(c) && (c != '_'))
            wrd.tok = TOK_ERROR;
			else {
	
				/* stop when reaching nonchar token or disallowed character */
				do {
					
					*word_buf_ptr++ = c;
					tok = read_token();
					if (!IS_TOK_CHAR(tok)) break;
					c = GET_TOK_CHAR(tok);
					
				} while (isalpha(c) || isdigit(c) || c == '_');
				
				/* cut the word, link it to wrd */
				*word_buf_ptr++ = '\0';
				wrd.tok = is_var;
				wrd.text = text_start;
				
				/* check presence of extended vars */
				if ((was_lt == TOK_LT && tok == TOK_GT) || 
					 (was_lt == TOK_LT2 && tok == TOK_GT2))
					/* properly bracketed */
					tok = read_token();
				else if (was_lt)
					/* no closing bracket right after varname end */
					wrd.tok = TOK_ERROR;
			}

			is_var = 0;
			old_token = tok;
			
		} else if ((is_var == TOK_DOLAR && tok == TOK_LT) ||
					 (is_var == TOK_DOLAR2 && tok == TOK_LT2)) {
		  /* $< pair */
		  was_lt = tok;
		  goto read_word_start;
		} else
		  /* nothing else allowed in variable */
			wrd.tok = TOK_ERROR;

	} else {

		if (IS_TOK_CHAR(tok)) {

		   /* gather characters into word */	
			do {
				*word_buf_ptr++ = GET_TOK_CHAR(tok);
				tok = read_token();
			} while (IS_TOK_CHAR(tok));
			
			/* cut the word, link it to wrd */
			*word_buf_ptr++ = '\0';
			wrd.text = text_start;
			wrd.tok = TOK_WORD;

         /* single '2' followed by '>' could be part of redirection */
			if (*text_start == '2' && word_buf_ptr - text_start == 2 &&
				 tok == TOK_GT) {
			  word_buf_ptr = text_start;
			  wrd.tok = TOK_TWO;
			}

			old_token = tok;

         was_word = 1;
			return wrd;
			
		} else if ((tok == TOK_DOLAR) || (tok == TOK_DOLAR2)) {
		   /* save dolar type */
		   is_var = tok;
		   was_lt = 0;
			was_word = 0;
		   goto read_word_start;
			
		} else if (tok == TOK_BLANK) {
         /* squish blanks */
			do {
				tok = read_token();
			} while (tok == TOK_BLANK);
				
			/* save nonblank token 4 l8r */
			old_token = tok;
			wrd.tok = TOK_BLANK;
			was_word = 0;

		} else if (tok == TOK_LT2 || tok == TOK_GT2) {
		  /* <> in dquote have normal meaning
			  redo as when reading ordinary characters */
		  old_token = (tok == TOK_LT2) ? MAKE_TOK_CHAR('<') : MAKE_TOK_CHAR('>');
		  was_word = 0;
		  goto read_word_start;

		} else if (tok == TOK_NONE) {
      /* only after blocking word concatenation, discard */
		   was_word = 0;
         goto read_word_start;
		} else { 
			wrd.tok = tok;
			was_word = 0;
		}
	}
	
   return wrd;
}


LOCAL int reading_var;          /* currently parsing variable value */
LOCAL char *var_value;          /* reference to current variable string */
LOCAL char var_buffer[2*(MAX_LINE + 1)]; /* buffer for var chars */
LOCAL char var_err_buffer[MAX_LINE + 1]; /* buffer for var chars */
LOCAL char *var_buf_ptr;                 /* current position in buffer */
LOCAL word old_wrd;                      /* old word token */

/** find value for a variable */
LOCAL char *ref_var(char *name)
{
   if (lex_state == LEX_CHECK) {
      strcpy(&var_err_buffer[1],name);
      return var_err_buffer;
   } else
      return get_variable(name);
}

/** initialize variable part of parser */
LOCAL int init_variable_parser(void)
{
   reading_var = 0;
	var_buf_ptr = var_buffer;
	old_wrd.tok = TOK_UNDEF;
   var_err_buffer[0] = '$';
   return 0;
}

/** cleanup for variable parser */
LOCAL int done_variable_parser(void)
{
   return 0;
}

/** expand variables for their values */
LOCAL word read_variable(void)
{
   word wrd;
	char c;
	char *word_start;

	word_start = var_buf_ptr;
   /* normal variable outside dquotes */
   if (reading_var == 1) {

		/* expand current variable into tokens */
      switch (*var_value) {
         case ' ':
			case '\t':
				/* squeeze spaces */
				while (*var_value == ' ' || *var_value == '\t')
					var_value++;
				wrd.tok = TOK_BLANK;
				break;
			case '*':
				/* create star */
				var_value++;
				wrd.tok = TOK_STAR;
				goto check_end;
			case '?':
			   /* create qmark */	
				var_value++;
				wrd.tok = TOK_QMARK;
check_end:
				/* at the end of variable expanded string of type
               $<var_name>2>word, we have to create a virtuall
					entry containing a single 2 character to prevent from
               joining into '2>', which would not be correct */

				if (*var_value == '\0') {
					old_wrd = read_word();
					if (old_wrd.tok == TOK_TWO) {
						old_wrd.tok = TOK_WORD;
						old_wrd.text = "2";
					}
				}
				
				break;
			default:
				/* eat other characters */
				while ((c = *var_value) && (c != ' ') && (c != '\t') &&
						 (c != '*') && (c != '?')) {
               *var_buf_ptr++ = c;
					var_value++;
				}
			   
				/* the same comment as above, different handling:
					glue the character 2 to the value of expanded variable */
				
		      if (*var_value == '\0') {
					old_wrd = read_word();
					if (old_wrd.tok == TOK_TWO) {
						*var_buf_ptr++ = '2';
						old_wrd.tok = TOK_UNDEF;
					} 
				}

				/* create word structure item, pointing into buf */
				*var_buf_ptr++ = '\0';
				wrd.text = word_start;
				wrd.tok = TOK_WORD;
				break;
      }

		if (*var_value == '\0') reading_var = 0; /* step out of var */

   } else {

     if (old_wrd.tok == TOK_UNDEF)
		  wrd = read_word();
	  else {
		  wrd = old_wrd;
        old_wrd.tok = TOK_UNDEF;
	  }

     if (wrd.tok == TOK_DOLAR) {
        reading_var = 1;
        var_value = ref_var(wrd.text);
        /* check for defined variable */
        if (!var_value) reading_var = 0;

        /* there was an empty variable expansion*/
		  wrd.tok = TOK_NOVAR;

     } else if (wrd.tok == TOK_DOLAR2) {

		 /* for variables in dquotes no wildcard expansion is allowed */
       var_value = ref_var(wrd.text);

		 if (!var_value)
			 wrd.tok = TOK_NOVAR;
		 else {
		 	wrd.text = var_value;
			wrd.tok = TOK_WORD;
		 }
	  }
   }
   return wrd;
}

LOCAL char glue_buf[MAX_LINE + 1];
LOCAL char expand_buf[MAX_LINE + 1];
LOCAL int in_wild;         /* in wildcard flag  */
LOCAL word old_var;
LOCAL char **exp_base;     /* base pointer for expanded string array */
LOCAL char **exp_curr;     /* walking pointer */

/** initialize wildcard expansion parser */
LOCAL int init_glue_parser(void)
{
   in_wild = 0;
	exp_base = NULL;
	old_var.tok = TOK_UNDEF;
   return 0;
}

/** cleanup wildcard expansion parser */
LOCAL int done_glue_parser(void)
{
	free_expand(); /* deallocate array of expanded filenames */
   return 0;
}

/** wildcar expanison parser */
LOCAL word read_glue(void)
{
   /* for checking phase and error messages */
	static char *dummy_base[2] = { "wildcard expansion" , NULL};
   word var;
	token t;
	int run_expand;
	char c, *p, *q, *r;
	
read_glue_start:
	
	/* in wildcard*/
   if (in_wild) {
		/* output word */
		var.tok = TOK_WORD;
		var.text = *exp_curr;
		exp_curr++;
			
      if (!*exp_curr)
			/* end of list of words */
			in_wild = 0;

	} else {
		
		if (old_var.tok == TOK_UNDEF) 
			var = read_variable();
		else {
         var = old_var;
			old_var.tok = TOK_UNDEF;
		}

		t = var.tok;
		run_expand = 0;
		
		/* glue wildcard word together */
		if (t == TOK_WORD || t == TOK_QMARK || t == TOK_STAR) {
			
			p = expand_buf;
			q = glue_buf;
			
         do {

				if (var.tok == TOK_WORD) {
					r = var.text;

					while ((c = *r++) != '\0') {
					/* prevent from treating ordinary characters as wildcards */
						 if (c == '\\' || c == '[' || c == '*' || c == '?') {
							 *p++ = '\\';
							 *p++ = c;
						} else 
						  *p++ = c;

						*q++ = c;
					}

				} else if (var.tok == TOK_STAR) {
				  run_expand = 1;
				  *p++ =  *q++ = '*';
				} else if (var.tok == TOK_QMARK) {
				  run_expand = 1;
				  *p++ = *q++ = '?';
				} else if (var.tok == TOK_TWO)
				  *p++ = *q++ = '2';

		      var = read_variable();
				t = var.tok;
				
				/* accept empty expansion as an empty part of a word */
			} while (t == TOK_WORD || t == TOK_QMARK || t == TOK_STAR || 
					   t == TOK_NOVAR || t == TOK_TWO);

         *p = '\0';
         *q = '\0';

         old_var = var;

         if (run_expand) {
				if (lex_state == LEX_CHECK) {
					/* feed in some virtual expansion */
					exp_curr = exp_base = dummy_base;
					in_wild = 1;
					goto read_glue_start;

				} else {

					free_expand();    /* for safety */

					/* perform expansion */
					if ((exp_base = expand(expand_buf)) != NULL) {
                  
						exp_curr = exp_base;
						in_wild = 1;

						goto read_glue_start;
					}
					/* fall through */
				}
			}
			
			var.text = glue_buf;
			var.tok = TOK_WORD;			
		}
   }
	
	return var;
}

/* enum constants have to be synced with bison %tokens in all bison files */
typedef enum {
  LEX_IN        = 300,   /* < */
  LEX_OUT       = 301,   /* > */ 
  LEX_2OUT      = 302,   /* 2> */ 
  LEX_APP       = 303,   /* >> */ 
  LEX_2APP      = 304,   /* 2>> */ 
  
  LEX_PIPE      = 305,   /* | */

  LEX_SEMICOLON = 306,   /* ; */
  LEX_OR        = 307,   /* || */
  LEX_AND       = 308,   /* && */
  
  LEX_EOLN      = 309,   /* end of line (end of input) */
  LEX_ERROR     = 310,   /* error flag */
	  
  LEX_WORD      = 311,   /* normal word */
  LEX_EQ        = 312    /* = */
} lex_type;

/* lex parser local states */
typedef enum {
  IN_UNDEF,	
  IN_OUT,        /* > */ 
  IN_2,          /* 2 */ 
  IN_2OUT,       /* 2> */ 
  IN_OR,         /* | */
  IN_ET          /* & */
} in_lex;


LOCAL word old_glue;
LOCAL int in_lex_token;
LOCAL in_lex lex_curr;
GLOBAL char *lex_word;            /* word to pass from lex */
LOCAL int should_eof;             /* eof is next */
LOCAL const char *emptyword = ""; /* constant empty word */

/** initialize lexical parser */
GLOBAL int init_lex_parser(char *line)
{
   if (init_token_parser(line) || init_word_parser() || 
	    init_variable_parser() || init_glue_parser()) return 1;

	old_glue.tok = TOK_UNDEF;
	in_lex_token = 0;
   should_eof = 0;

   return 0;
}

/** finalize lexical parser */
GLOBAL int done_lex_parser(void)
{
	if (done_glue_parser() || done_variable_parser() || 
		 done_word_parser() || done_token_parser()) return 1;
   return 0;
}


/** read lexical token 
    corresponds to yy_lex 
    returns LEX_??? constants
*/
LOCAL lex_type read_lex_aux(void)
{
   word w;
	token t;

read_lex_start:
	
	if (old_glue.tok == TOK_UNDEF) 
      w = read_glue();
	else {
		w = old_glue;
		old_glue.tok = TOK_UNDEF; 
	}

	t = w.tok;
	
   if (in_lex_token) {

      switch (lex_curr) {
			case IN_OUT:
				in_lex_token = 0;
				if (t == TOK_GT)
               return LEX_APP;
				old_glue = w;
				return LEX_OUT;
			case IN_2:
				/* t == TOK_GT */
				lex_curr = IN_2OUT;
				goto read_lex_start;
			case IN_2OUT:
				in_lex_token = 0;
				if (t == TOK_GT) 
					return LEX_2APP;
				old_glue = w;
				return LEX_2OUT;
			case IN_OR:
				in_lex_token = 0;
				if (t == TOK_OR) 
					return LEX_OR;
				else {
					old_glue = w;
					return LEX_PIPE;
				}
			case IN_ET:
				in_lex_token = 0;
				if (t == TOK_ET)
					return LEX_AND;
				else
					return LEX_ERROR;
			default:
				/* should not happen */
				break;
		}
		
   } else {
		
		switch (t) {
			case TOK_GT:
				/* out redirections */
				in_lex_token = 1;
				lex_curr = IN_OUT;
				goto read_lex_start;
			case TOK_TWO:
				/* part of 2> redirection */
				in_lex_token = 1;
				lex_curr = IN_2;
				goto read_lex_start;
			case TOK_OR:
				/* part of | or || */
				in_lex_token = 1;
				lex_curr = IN_OR;
				goto read_lex_start;
			case TOK_ET:
				/* part of && */
				in_lex_token = 1;
				lex_curr = IN_ET;
				goto read_lex_start;
			case TOK_LT:
				/* < redirection */
				return LEX_IN;
			case TOK_SEMICOLON:
				return LEX_SEMICOLON;
			case TOK_BLANK:
			case TOK_NOVAR:
				/* discard blanks and empty expansions */
            goto read_lex_start; 
			case TOK_EOLN:
				return LEX_EOLN;
			case TOK_ERROR:
				return LEX_ERROR;
			case TOK_EQ:
				return LEX_EQ;
			case TOK_WORD:
				lex_word = w.text;
				return LEX_WORD;
			case TOK_EMPTY:
            /* empty word as "" or '' */
            lex_word = (char *)emptyword;
            return LEX_WORD;
			default:
				/* should not happen */
				break;
		}
	}
	
   return 0;
}

/** public interface to lexical analyser */
GLOBAL int read_lex(void)
{
   int i;
   if (should_eof) return 0; /* eof */

   i = (int)read_lex_aux();	
   /* next time signal EOF to cut syntax analysis */
   if (i == LEX_EOLN) should_eof = 1;
   return i;
}


#if defined(TEST_LEXER1) || defined(TEST_LEXER2)

#include <stdio.h>

LOCAL char *read_line(char *s, int max)
{
	int l;
	if (!fgets(s,max,stdin)) return NULL;

	l = strlen(s);
	if (l && s[l - 1] == '\n') s[l - 1] = '\0';

   return s;
}

#endif

#ifdef TEST_LEXER1
char reverse[]= "c$@<{>};|&?*=_nE0";

int parse_line(char *line)
{
   token tok;

   if (init_token_parser(line)) return 1;

   do {
		tok = read_token();
      if (IS_TOK_CHAR(tok)) 
         printf("char token: %c\n",GET_TOK_CHAR(tok));
		else 
			printf("special token: %c\n",reverse[tok]);

	} while (tok != TOK_ERROR && tok != TOK_EOLN);
	
	if (done_token_parser()) return 1;
	
   return tok == TOK_ERROR;
}

int main(void)
{
    char line[MAX_LINE + 1];

    do {
		 if (!read_line(line,MAX_LINE)) break;
		 init_lex_parser(line);
		 if (parse_line(line)) printf("parser error!\n");
		 done_lex_parser();
    } while (1);

    return EXIT_SUCCESS;
}

#endif



#ifdef TEST_LEXER2
LOCAL char *lex_names[] = { "<",">","2>",">>","2>>","|",";","||",
"&&","EOLN","error","word","=" };


int parse_line(char *line)
{
   int lex;

   do {
		lex = read_lex();
		
      if (lex == LEX_WORD) 
         printf("word token: '%s'\n",lex_word);
		else if (lex == 0)
           printf("error in lex!\n");
     else
			printf("special token: %s\n",lex_names[lex - 300]);

	} while (lex != LEX_ERROR && lex != LEX_EOLN);
	
   return 0;
}

extern char **environ;

int main(void)
{
    char line[MAX_LINE + 1];

    lex_state = LEX_BUILD;
/*    lex_state = LEX_CHECK;*/

    init_vars(environ);

    do {
		 if (!read_line(line,MAX_LINE)) break;
		 init_lex_parser(line);
		 if (parse_line(line)) printf("parser error!\n");
		 done_lex_parser();
    } while (1);

    done_vars();

    return EXIT_SUCCESS;
}
#endif

