/**
     run.c
 
     process running
 
     Petr Zika 2002
*/
           
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include "../comdefs.h"
#include "../errors/errors.h"
#include "../utils/utils.h"
#include "../env/env.h"
#include "../tree/tree.h"
#include "../io/io.h"
#include "../main/main.h"


/* default path specification */
LOCAL char *default_path = "PATH=/bin:/usr/bin";

/** perform file execution like execve, search path */
LOCAL int execvep(char *filename, char **argv, char **envp)
{
   char **e;
	char c;
	char *p, *q, *r, *rend;
	int i, m, n, rlen;
	int flag;

   clear_handler();
	
	/* do not search PATH if '/' is a part of filename */
   if (strchr(filename,'/') != NULL) {
		argv[0] = filename;   /* replace with command name */
		execve(filename,argv,envp);
		/* we get here unless the command is executed */
		reset_handler();
		return -1; 
	}
	
	flag = 1;
   e = envp;
   while (*e) {
		if (!strncmp(*e,"PATH=",5)) {
			flag = 0;
			break;
		}
		e++;
   }

	if (flag) p = default_path; else p = *e;
   p += 5; /* skip PATH=*/

   /* hack: the buffer is filled from back to front */
   
	/* find the longest string among path list */
	q = p;
	i = n = 0;
	
	while ((c = *q) != '\0') {
		if (c == ':') {
			if (i > n) n = i;
			i = -1;
		}
		i++;
		q++;
	}

	if (i > n) n = i; /* last entry */
	
   m = strlen(filename);
	/* rlen is enough for the longest path, '/' filename and '\0' */
	rlen = n + m + 2;
	
   r = (char *)MALLOC(sizeof(char)*rlen);
   
	/* set rend to the beginning of filename space in string r */
	rend = r + (rlen - 1) - m;
   /* place filename to the end of the buffer */
   strcpy(rend,filename);
   /* add an usefull slash to separate the filename */
	*(--rend) = '/';

	while (*p) {
		/* find the end of current path string */
		q = strchr(p,':'); 
		if (!q) q = strchr(p,'\0');
		n = (q - p) / sizeof(char);

		/* glue path and filename together */
		strncpy(rend - n,p,n);

		p += n;
		if (*p == ':') p++;  /* if not at the end */

		argv[0] = rend - n;
		/* attempt ... */
		execve(rend - n,argv,envp);
   }
	
	free(r);
	reset_handler();
	return -1;  /* return the last error ... */
}

LOCAL char tmp_path[PATH_MAX + 1];

/** change directory within shell, update PWD */
LOCAL int change_dir(char *newdir)
{
    
	 if (chdir(newdir) == -1) {
		 perror("cd");
		 return 1;
	 }

   if (getcwd(tmp_path,PATH_MAX + 1) == NULL) {
		perror("cwd");
		return 1;
	}
	 
	if (set_variable("PWD",tmp_path)) {
		error("cannot change environment");
		return 1;
	}
	return 0;
}

LOCAL int run_commands(cmd_list *c)
{
	pid_t pid, last_pid;
	int fd;      /* temporary descriptor */
	int fd_pipe[2]; /* in/out fd through pipe */
	int in_pipe;    /* input fd from the last process */
	int i, nrunning, attr;
	int status, last_status;
	cmd_list *d;
   char **argv;

	in_pipe = -1;
	last_pid = -1;
	nrunning = 0;
   d = c;
	while (d) {

		if (d->next) {
			if (pipe(fd_pipe) == -1) {
				perror("creating pipe");
				for (i = 0; i < nrunning; i++)
					 wait(&status);
				return -1;
			}
		} else {
			fd_pipe[0] = -1;  /* set to undefined to prevent from closing */
			fd_pipe[1] = -1;
		}

		if ((pid = fork()) == -1) {
         /* error encountered, still in parent */
	      perror("process fork");
			for (i = 0; i < nrunning; i++)
				 wait(&status);
			if (d->next) {
			   if (close(fd_pipe[0]) == -1) perror("close");
			   if (close(fd_pipe[1]) == -1) perror("close");
			}
				
			return -1;
		} else if (pid == 0) {
			/* child process */

         if (in_pipe != -1) {
				/* attach last output to stdin */
				if (dup2(in_pipe,0) == -1) {
					perror("dup2");
					exit(1);
				}
				/* old descriptor not needed anymore */
				if (close(in_pipe) == -1) {
					perror("close");
					exit(1);
				}
			}
			
			/* just in case there is a pipe opened */
			if (fd_pipe[0] != -1) {
			   /* close input side of descriptor */
				if (close(fd_pipe[0]) == -1) {
					perror("close");
					exit(1);
				}
				/* attach output side of descriptor to stdout */
				if (dup2(fd_pipe[1],1) == -1) {
					perror("dup2");
					exit(1);
				}
				/* get rid of the old descriptor */
				if (close(fd_pipe[1]) == -1) {
					perror("close");
					exit(1);
				}
         }

			/* some input redirection */
			if (d->in.type != RED_NONE) {
				if ((fd = open(d->in.name,O_RDONLY)) == -1) {
					perror(d->in.name);
					exit(1);
				}
				if (dup2(fd,0) == -1) {
					perror("dup2");
					exit(1);
				}
				if (close(fd) == -1) {
					perror("close");
					exit(1);
				}
			}
			
			/* output redirection */
			if ((i = d->out.type) != RED_NONE) {
				if (i == RED_OUT)
					attr = O_WRONLY | O_CREAT | O_TRUNC;
				else
					/* i == RED_APP */
					attr = O_WRONLY | O_CREAT | O_APPEND;
				
				if ((fd = open(d->out.name,attr,0666)) == -1) {
					perror(d->out.name);
					exit(1);
				}
				if (dup2(fd,1) == -1) {
					perror("dup2");
					exit(1);
				}
				if (close(fd) == -1) {
					perror("close");
					exit(1);
				}
			}

			/* error redirection */
			if ((i = d->err.type) != RED_NONE) {
				if (i == RED_2OUT) 
					attr = O_WRONLY | O_CREAT | O_TRUNC;
				else
					/* i == RED_2APP */
					attr = O_WRONLY | O_CREAT | O_APPEND;
				
				if ((fd = open(d->err.name,attr,0666)) == -1) {
					perror(d->err.name);
					exit(1);
				}
				if (dup2(fd,2) == -1) {
					perror("dup2");
					exit(1);
				}
				if (close(fd) == -1) {
					perror("close");
					exit(1);
				}
			}

			if ((argv = create_argv(d->args)) == NULL) {
					error("memory error");
					exit(1);
			}

			execvep(d->name,argv,get_env());
			
			/* if we get here, sure the execution failed */
			perror("exec");
			error("error executing `%s'",d->name);
			exit(1);
		} else {
			/* parent continuing */

			last_pid = pid; /* save child pid */
			if (in_pipe != -1) {
				if (close(in_pipe) == -1) {
					perror("close");
					for (i = 0; i < nrunning; i++)
						 wait(&status);
					return -1;
				}
			}

         if (fd_pipe[1] != -1) {
				if (close(fd_pipe[1]) == -1) {
					perror("close");
					for (i = 0; i < nrunning; i++)
						 wait(&status);
					return -1;
				}
			}

			/* save input descriptor for next process */
			in_pipe = fd_pipe[0];
			nrunning++;
		}

		d = d->next;
	}

	for (i = 0; i < nrunning; i++) {
		 while ((pid = wait(&status)) == -1)
			 perror("wait");

		 if (pid == last_pid)
			 last_status = status;
	}

	/* find the last process' return code */
	if (WIFEXITED(last_status))
    	return WEXITSTATUS(last_status);
   else if(WIFSIGNALED(last_status))
		return 1; /* return something != 0 */
	else {
		error("odd process status");
		return -1;
	}
}

/** run all commands, recognize internal commands */
GLOBAL int run_cmd_list(cmd_list *c)
{
	char *name;
	char *arg;
	char *s;
	int run = 0;

   name = c->name;
	
   /* process internal commands first */
	if (!strcmp(name,"cd")) {
		if (c->args == NULL) {
			if ((s = get_variable("PWD")) == NULL) {
				error("cannot read environment");
				return -1;
			}
			printf("%s\n",s);
		} else {
			if (change_dir(arg = c->args->name)) {
				error("cannot chdir to `%s'",arg);
				return -1;
			}
		}
	} else if (!strcmp(name,"export")) {
		if (c->args == NULL)
			print_env();
		else {
			if (export_variable(arg = c->args->name)) {
				error("cannot export `%s'",arg);
				return -1;
			}
		}
		
	} else if (!strcmp(name,"local")) {
		if (c->args == NULL) {
			print_local();
		} else {
			if (localize_variable(arg = c->args->name)) {
				error("cannot make `%s' local",arg);
				return -1;
			}
		}
	} else if (!strcmp(name,"unset")) {
		if (c->args == NULL) {
			error("nothing to unset");
		} else {
			if (unset_variable(arg = c->args->name)) {
				error("cannot unset `%s'",arg);
				return -1;
			}
		}
	} else if (!strcmp(name,"exit")) {
		/* signal program exit */
		return -2;
	} else run = 1;

   if (!run) {
		if (c->next || c->in.type || c->out.type || c->err.type) 
			error("pipe or redirection not allowed for internal `%s'",c->name);
		return -1;
   }

	return run_commands(c);
}

#ifdef TEST_RUN

int main(void)
{
   char *args[] = { NULL };
	char *env[] = { NULL, "abc=def", "PATH=/tmp/dummy:/tmp/dummy1:/tmp/dummy2:/tmp/dummy3:/usr/bin", NULL };
	int ret;
	
	ret = execvep("nothing",args,env);
   if (ret == -1) error("system problem with nothing: %s",strerror(errno));

	ret = execvep("ls",args,env);
   if (ret == -1) error("system problem with ls: %s",strerror(errno));

	return 0;
}

#endif

