/**
       tree.c
 
       parsed commands storage structures
 
       Petr Zika 2002
*/

#define _XOPEN_SOURCE 500

#include <features.h>
#include <string.h>
#include <stdlib.h>
#include "../comdefs.h"
#include "tree.h"
#include "../errors/errors.h"
#include "../io/io.h"

/* GLOBAL int tree_fd = stdout; */

/* print argument list*/
GLOBAL void print_arg_list(arg_list *a)
{
   while (a) {
		printf(" %s",a->name);
		a = a->next;
   }
}

/** print redirection list */
GLOBAL void print_redir_list(redir_list *r)
{
   while (r) {
	   if (r->type == RED_NONE) continue;
   	putchar(' ');
	   switch (r->type) {
			case RED_IN:
				putchar('<');
				break;
			case RED_2OUT:
				putchar('2');
				putchar('>');
				break;
		   /* warning! this is a fall through hack */
			case RED_2APP:
				putchar('2');
			case RED_APP:
				putchar('>');
			case RED_OUT:
				putchar('>');
				break;
			default:
			   fatal("print_redir wrong type");
				break;
		}
		printf("%s",r->name);

		r = r->next;
   }
}


GLOBAL void print_redir(redir r)
{
   if (r.type == RED_NONE) return;

   putchar(' ');
	switch (r.type) {
		case RED_IN:
			putchar('<');
			break;
		case RED_2OUT:
			putchar('2');
			putchar('>');
			break;
		case RED_2APP:
			putchar('2');
		case RED_APP:
			putchar('>');
		case RED_OUT:
			putchar('>');
			break;
		default:
			fatal("print_redir wrong type");
			break;
	}
	printf("%s",r.name);
}

/** print command list */
GLOBAL void print_cmd_list(cmd_list *c)
{
	int next = 0;

   while (c) {
		if (next) printf(" | "); else next = 1;
		printf("%s",c->name);
		print_arg_list(c->args);
		print_redir_list(c->redirs);
      print_redir(c->in);
      print_redir(c->out);
      print_redir(c->err);
		c = c->next;
   }

/* XXX putchar('\n');  perhaps they do not want it formatted .. */
}

/** create argument array for exec, leave name empty */
GLOBAL char **create_argv(arg_list *a)
{
   arg_list *b;
	char **p, **q;
	char *s;
	int n, len;

	n = 2;    /* name && last NULL */
	len = 0;
	b = a;
	
	while (b) {
		len += strlen(b->name) + 1;
		n++;
		b = b->next;		
	}
	
	if ((p = (char **)malloc(sizeof(char *)*n)) == NULL)
		return NULL;

	if (len) {
		if ((s = (char *)malloc(sizeof(char)*len)) == NULL) {
			free(p);
			return NULL;
		}
   }

	b = a;
	q = p + 1;  /* skip argv[0] */
   n--;
	
	/* fill both the array and the buffer at once */
	while (--n) {
		strcpy(s,b->name);
		*q = s;
		s = strchr(s,'\0') + 1; /* find the end of the string */
		b = b->next;
		q++;
	}

   *q = NULL;
	
   return p;
}

/* print argv structure, possibly argv[0] is zero */
GLOBAL void print_argv(char **a)
{
	int next = 0;

   if (!*a) a++;	
	while (*a) {
		if (next) putchar(' '); else next = 1;
		printf("%s",*a);
		a++;
	}
}

/* free previously created argv structures */
GLOBAL void free_argv(char **a)
{
   free(*a);  /* free the first entry, actually the whole block of strings */
	free(a);   /* free the array */
}

/** prune redirection list */
GLOBAL void prune_redir_list(cmd_list *c, int pipe_out)
{
   redir_list *r = c->redirs;
   redir_list *s;
	redir_list *in, *out, *err;

   c->in.type = c->out.type = c->err.type = RED_NONE;
	in = out = err = NULL;
	
	s = r;
   while (s) {
		switch (s->type) {
			case 	RED_IN:
				in = s;
				break;
			case 	RED_OUT:
			case 	RED_APP:
				out = s;
				break;
			case 	RED_2OUT:
			case 	RED_2APP:
				err = s;
				break;
		}
		s = s->next;
   }

	if (in) {
		c->in.name = in->name;
		c->in.type = in->type;
		in->name = NULL;     /* detach string from previous structure */
	}

	/* only in case there is no pipe overriding redirection */
	if (out && !pipe_out) {
		c->out.name = out->name;
		c->out.type = out->type;
		out->name = NULL;     
	}
	
	if (err) {
		c->err.name = err->name;
		c->err.type = err->type;
		err->name = NULL;     
	}

	free_redir_list(r);  /* release all unused */
   c->redirs = NULL;
}
/** remove ambiguities in the whole command list */
GLOBAL void prune_cmd_list(cmd_list *c)
{
   while (c) {
		prune_redir_list(c,(c->next != NULL));
		c = c->next;
   }
}

/** free argument list */
GLOBAL void free_arg_list(arg_list *a)
{
   arg_list *b;

   while (a) {
		b = a->next;
		free(a->name);
		free(a);
		a = b;
   }
}

/** free redirection list */
GLOBAL void free_redir_list(redir_list *r)
{
   redir_list *s;

   while (r) {
		s = r->next;
		free(r->name);
		free(r);
		r = s;
   }
}

/** free command list */
GLOBAL void free_cmd_list(cmd_list *c)
{
   cmd_list *d;

   while (c) {
		d = c->next;
		free_arg_list(c->args);
		free_redir_list(c->redirs);
		free(c->name);
		free(c);
		c = d;
   }
}

#ifdef TEST_TREE

int main(void)
{
   arg_list *a, *b;
	char **ar;
	char *c;
	int i;

	a = NULL;
	
	for (i = 0; i < 10; i++) {
		
		b = (arg_list *)malloc(sizeof(arg_list));
		c = (char *)malloc(sizeof(char)*21);
		snprintf(c,20,"test %d",i);
		b->name = c;
		b->next = a;
		a = b;
	}
	
	ar = create_argv("tree",a);
	print_argv(ar);
   putchar('\n');
	
	free_argv(ar);
	free_arg_list(a);
	
	return 0;
}

#endif

